﻿#include "RTU_Salve.h"
#include <bitset>
char read_buf[MAX_NUMBER];

bitset<MAX_Address> Coil_date; //0-9999
UINT16 Register[MAX_Address]; // 0-9999

/*********************************************************************************************
* 功能     ：   初始化线圈和寄存器
* 描述	   ：	对线圈和寄存器数组赋值为全1
* 输入     ：   无
* 返回值   ：   无
**********************************************************************************************/
void Init_Coil_Register(void)
{
	for (int i = 0; i < MAX_Address; i++)
	{
		Coil_date[i] = 1;
		Register[i] = 0xFFFF;
	}
}
/*********************************************************************************************
* 功能     ：   Bitset转UINT8类型
* 描述	   ：	Bitset ==》UINT8
* 输入     ：   Bitset_Address Bitset的起始地址  Read_Number 要读取的位数
* 返回值   ：   Date 转换后的UNIT8数据
**********************************************************************************************/
UINT8 Bitset_to_Uint8(unsigned int Bitset_Address, unsigned int Read_Number)
{
	UINT8 Date = 0x00;
	
	if (Read_Number >= 8)
	{
		unsigned int len = Bitset_Address + 8;
		for (int i = 0; i < 8; i++)
		{
			Date = Date << 1 | (int)Coil_date[len--];
		}
	}
	else
	{
		unsigned int len = Bitset_Address + Read_Number-1;
		for (int i = 0; i < Read_Number; i++)
		{
			Date = Date << 1 | (int)Coil_date[len--];
		}
	}
	return Date;
}

/*********************************************************************************************
* 功能     ：   生成0x01功能码响应报文
* 描述	   ：	通过判断地址范围生成0x01对应的响应报文
* 输入     ：   *Requst_Message 请求报文 *Response_Message响应报文
* 返回值   ：   无
**********************************************************************************************/
void Create_0x01_Response_Message(UINT8 *Requst_Message, UINT8 *Response_Message)
{ 
	unsigned int Start_Address = Requst_Message[2] << 8 | Requst_Message[3];
	unsigned int Address_range = Start_Address + (Requst_Message[4] << 8 | Requst_Message[5]);
	if (Address_range < MAX_Address)//判断地址是否超限
	{
		unsigned int Read_Number = Requst_Message[4] << 8 | Requst_Message[5];//要读取的位数
		unsigned int Read_Len = Count_Read_date_number(Requst_Message[0], Read_Number); //要读取的字节数
		for (unsigned int i = 3; i < Read_Len + 3; i++)
		{
			Response_Message[i] = Bitset_to_Uint8(Start_Address, Read_Number);
			Start_Address += 8;
			Read_Number -= 8;
		}
		Response_Message[0] = Device_ID;
		Response_Message[1] = Requst_Message[1];
		Response_Message[2] = Read_Len+2;  //响应报文中的后续字节数
	}
	else //地址超限
	{
		Response_Message[0] = Device_ID;
		Response_Message[1] = Requst_Message[1] + 0x80;
		Response_Message[2] = 0x02;
	}
}

/*********************************************************************************************
* 功能     ：   生成0x03功能码响应报文
* 描述	   ：	通过判断地址范围生成0x03对应的响应报文
* 输入     ：   *Requst_Message 请求报文 *Response_Message响应报文
* 返回值   ：   无
**********************************************************************************************/
void Create_0x03_Response_Message(UINT8 *Requst_Message, UINT8 *Response_Message)
{
	unsigned int Start_Address = Requst_Message[2] << 8 | Requst_Message[3];
	unsigned int Address_range = Start_Address + (Requst_Message[4] << 8 | Requst_Message[5]);
	if (Address_range < MAX_Address) //判断地址是否超限
	{
		unsigned int Read_Number = Requst_Message[4] << 8 | Requst_Message[5];//要读取的寄存器数量
		unsigned int Read_Len = Count_Read_date_number(Requst_Message[0], Read_Number); //要读取的字节数
		for (unsigned int i = 3; i < Read_Len + 3; i= i + 2)
		{
			Response_Message[i] = Register[Start_Address] >> 8;
			Response_Message[i + 1] = (UINT8)Register[Start_Address];
			Start_Address += 1;
		}
		Response_Message[0] = Device_ID;
		Response_Message[1] = Requst_Message[1];
		Response_Message[2] = Read_Len + 2; //响应报文中的后续字节数
	}
	else //地址超限 
	{
		Response_Message[0] = Device_ID;
		Response_Message[1] = Requst_Message[1] + 0x80;
		Response_Message[2] = 0x02;
	}
}

/*********************************************************************************************
* 功能     ：   写入线圈
* 描述	   ：	将数据写入线圈中
* 输入     ：   Write_date_Message 要写入的数据 Write_Number要写入的位数  Start_Address起始地址
* 返回值   ：   无
**********************************************************************************************/
void Write_Coil_date(UINT8 Write_date_Message, unsigned int Write_Number, unsigned int Start_Address)
{
	if (Write_Number >= 8)
	{
		for (unsigned int i = Start_Address; i < Start_Address+8; i++)
		{
			Coil_date[i] = Write_date_Message & 1;
			Write_date_Message >> 1;
		}
	}
	else
	{
		for (unsigned int i = Start_Address; i < Start_Address + Write_Number; i++)
		{
			Coil_date[i] = Write_date_Message & 1;
			Write_date_Message >> 1;
		}
	}
}

/*********************************************************************************************
* 功能     ：   生成0x0F功能码响应报文
* 描述	   ：	通过判断地址范围生成0x0F对应的响应报文
* 输入     ：   *Requst_Message 请求报文 *Response_Message响应报文
* 返回值   ：   无
**********************************************************************************************/
void Create_0x0F_Response_Message(UINT8 *Requst_Message, UINT8 *Response_Message)
{
	unsigned int Start_Address = Requst_Message[2] << 8 | Requst_Message[3];
	unsigned int Address_range = Start_Address + (Requst_Message[4] << 8 | Requst_Message[5]);
	if (Address_range < MAX_Address)//判断地址是否超限
	{
		unsigned int Write_Number = Requst_Message[4] << 8 | Requst_Message[5];//要写入的位数
		for (unsigned int i = 7; i < Requst_Message[6]+7; i++)//执行写入线圈操作
		{
			Write_Coil_date(Requst_Message[i], Write_Number, Start_Address);
			Write_Number -= 8;
			Start_Address += 8;
		}
		Response_Message[0] = Device_ID;
		Response_Message[1] = Requst_Message[1];
		Response_Message[2] = Requst_Message[2];
		Response_Message[3] = Requst_Message[3];
		Response_Message[4] = Requst_Message[4];
		Response_Message[5] = Requst_Message[5];
	}
	else //地址超限
	{
		Response_Message[0] = Device_ID;
		Response_Message[1] = Requst_Message[1] + 0x80;
		Response_Message[2] = 0x02;
	}
}

/*********************************************************************************************
* 功能     ：   生成0x10功能码响应报文
* 描述	   ：	通过判断地址范围生成0x10对应的响应报文
* 输入     ：   *Requst_Message 请求报文 *Response_Message响应报文
* 返回值   ：   无
**********************************************************************************************/
void Create_0x10_Response_Message(UINT8 *Requst_Message, UINT8 *Response_Message)
{
	unsigned int Start_Address = Requst_Message[2] << 8 | Requst_Message[3];
	unsigned int Address_range = Start_Address + (Requst_Message[4] << 8 | Requst_Message[5]);
	if (Address_range < MAX_Address)//判断地址是否超限
	{
		unsigned int Write_Number = Requst_Message[4] << 8 | Requst_Message[5];//要写入的寄存器个数
		for (unsigned int i = 7; i < Requst_Message[6] + 7; i = i + 2)//执行写入寄存器操作
		{
			Register[Start_Address] = Requst_Message[i] << 8 | Requst_Message[i+1];
			Start_Address++;
		}
		Response_Message[0] = Device_ID;
		Response_Message[1] = Requst_Message[1];
		Response_Message[2] = Requst_Message[2];
		Response_Message[3] = Requst_Message[3];
		Response_Message[4] = Requst_Message[4];
		Response_Message[5] = Requst_Message[5];
	}
	else //地址超限
	{
		Response_Message[0] = Device_ID;
		Response_Message[1] = Requst_Message[1] + 0x80;
		Response_Message[2] = 0x02;
	}
}

/*********************************************************************************************
* 功能     ：   生成异常功能码响应报文
* 描述	   ：	对不支持的功能码生成对应的异常响应报文
* 输入     ：   *Requst_Message 请求报文 *Response_Message响应报文
* 返回值   ：   无
**********************************************************************************************/
void Create_Abnormal_Function_Code_Response_Message(UINT8 *Requst_Message, UINT8 *Response_Message)
{
	Response_Message[0] = Device_ID;
	Response_Message[1] = Requst_Message[1] + 0x80;
	Response_Message[2] = 0x01;
}
/*********************************************************************************************
* 功能     ：   生成响应报文
* 描述	   ：	检查设备请求报文来生成对应功能的响应报文
* 输入     ：   *Requst_Message 请求报文 *Response_Message响应报文
* 返回值   ：   true  生成响应报文
*				false 不生成响应报文
**********************************************************************************************/
bool Create_Response_Message(UINT8 *Requst_Message, UINT8 *Response_Message)
{
	if (Requst_Message[0] != Device_ID) //检查设备ID一致
		return false;
	
	switch (Requst_Message[1])
	{
	case 0x01:Create_0x01_Response_Message(Requst_Message, Response_Message); break;
	case 0x03:Create_0x03_Response_Message(Requst_Message, Response_Message); break;
	case 0x0F:Create_0x0F_Response_Message(Requst_Message, Response_Message); break;
	case 0x10:Create_0x10_Response_Message(Requst_Message, Response_Message); break;
	default: Create_Abnormal_Function_Code_Response_Message(Requst_Message, Response_Message);
	}
	return true;
}



int test(void)
{
	string COMM;
	cin >> COMM;
	HANDLE Handle_Com = Init_COM((LPCTSTR)COMM.c_str(), 9600, 8, 0, 1);
	if (Handle_Com == INVALID_HANDLE_VALUE)
	{
		cout << "初始化串口失败" << endl;
		getchar();
		return 0;
	}

	char write_buf[MAX_NUMBER];
	memset(write_buf, 0, MAX_NUMBER);
	DWORD dwRead;
	while (true)
	{
		//RTU主站，生成并发送请求，计时，读取响应报文,先发送后接收
		//生成请求报文
		//发送请求报文

		PurgeComm(Handle_Com, PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR | PURGE_TXABORT);//清除缓存
		while (1)
		{
			if (true == SendData(Handle_Com, "ok", 2))
			{
				cout << "发送请求报文成功" << endl;
				break;
			}
		}
		BOOL bReadOK = ReadFile(Handle_Com, read_buf, 256, &dwRead, NULL);
		if (bReadOK && (dwRead > 0))
		{
			read_buf[dwRead] = '\0';
			printf("收到响应：%s \r\n", read_buf);
		}
		else
			cout << "loss" << endl;
	}
	CloseHandle(Handle_Com);
	getchar();
	return 0;
}
